__description__ = "a module that houses routines to automate instantiation of objects"
__author__ = "Reed Essick (reed.essick@ligo.org), Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

import importlib.util
import itertools
import os

import numpy as np
import pluggy

### non-standard libraries
from ligo.segments import segment, segmentlist

from . import classifiers
from . import features
from . import utils
from .io import reporters
from .io import triggers


#-------------------------------------------------

DEFAULT_DT = classifiers.DEFAULT_DT
DEFAULT_AVG_RATE_THRESHOLD = 1 ### feature vector density threshold to determine whether to build a regular or dense quiver

#-------------------------------------------------
# things from io.triggers

ClassifierDataFactory = triggers.ClassifierDataFactory ### This is declared in io so that we can use it within MultiSourceClassifierData

#------------------------


class ReporterFactory(object):
    """
    A factory to create instances of Reporter-like objects
    """
    def __init__(self):

        ### set up plugin manager
        manager = pluggy.PluginManager('iDQ')
        manager.add_hookspecs(reporters)

        ### load in reporters by module
        ### FIXME: do this with importlib instead when python3 only

        ### base reporters
        from .io.reporters import pkl, hdf5
        manager.register(pkl)
        manager.register(hdf5)

        ### extra (optional) reporters based on libraries
        try:
            from .io.reporters import gwf
        except ImportError:
            pass
        else:
            manager.register(gwf)

        try:
            from .io.reporters import ligolw
        except ImportError:
            pass
        else:
            manager.register(ligolw)

        try:
            from .io.reporters import kafka
        except ImportError:
            pass
        else:
            manager.register(kafka)

        ### add all registered plugins to registry
        self._reporters = {}
        for reporter_plugins in manager.hook.get_reporters():
            self._reporters.update(reporter_plugins)

    def __call__(self, rootdir, start, end, flavor=None, **kwargs):
        """
        standarize how we construct reporters
        """
        if flavor not in self._reporters:
            raise KeyError('%s is not a known Reporter!' %flavor) ### format error reporting a bit better
        return self._reporters[flavor](rootdir, start, end, **kwargs)

#-------------------------------------------------
# things from classifiers

class ClassifierFactory(object):
    """
    A factory to create instances of SupervisedClassifier-like objects
    """
    def __init__(self):

        ### set up plugin manager
        manager = pluggy.PluginManager('iDQ')
        manager.add_hookspecs(classifiers)

        ### load in classifiers by module
        ### FIXME: do this with importlib instead when python3 only

        ### base reporters
        from .classifiers import ovl
        manager.register(ovl)

        ### extra (optional) classifiers based on libraries
        try:
            from .classifiers import sklearn
        except ImportError:
            pass
        else:
            manager.register(sklearn)

        try:
            from .classifiers import keras
        except ImportError:
            pass
        else:
            manager.register(keras)

        try:
            from .classifiers import xgb
        except ImportError:
            pass
        else:
            manager.register(xgb)

        ### resolve local classifiers path
        suffix = os.path.join("idq", "classifiers.py")
        if "XDG_CONFIG_HOME" in os.environ:
            local_path = os.path.join(os.getenv("XDG_CONFIG_HOME"), suffix)
        else:
            local_path = os.path.join(os.getenv("HOME"), ".config", suffix)

        ### load local classifiers if available
        if os.path.exists(local_path):
            spec = importlib.util.spec_from_file_location("classifiers", local_path)
            local_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(local_module)
            manager.register(local_module)

        ### add all registered plugins to registry
        self._classifiers = {}
        for classifier_plugins in manager.hook.get_classifiers():
            self._classifiers.update(classifier_plugins)

    def __call__(self, nickname, flavor=None, **kwargs):
        if flavor not in self._classifiers:
            raise KeyError('%s is not a known SupervisedClassifier!'%flavor) ### format error reporting a bit more nicely
        return self._classifiers[flavor](nickname, **kwargs)


class IncrementalClassifierFactory(object):
    """
    A factory to create instances of IncrementalSupervisedClassifier-like objects
    """
    def __init__(self):

        ### set up plugin manager
        manager = pluggy.PluginManager('iDQ')
        manager.add_hookspecs(classifiers)

        ### load in classifiers by module
        ### FIXME: do this with importlib instead when python3 only

        ### base reporters
        from .classifiers import ovl
        manager.register(ovl)

        ### extra (optional) classifiers based on libraries
        try:
            from .classifiers import sklearn
        except ImportError:
            pass
        else:
            manager.register(sklearn)

        ### add all registered plugins to registry
        self._classifiers = {}
        for classifier_plugins in manager.hook.get_incremental_classifiers():
            self._classifiers.update(classifier_plugins)

    def __call__(self, nickname, flavor=None, **kwargs):
        if flavor not in self._classifiers:
            raise KeyError('%s is not a known IncrementalSupervisedClassifier!'%flavor) ### format error reporting
        return self._classifiers[flavor](nickname, **kwargs)


#-------------------------------------------------
# things from features

class QuiverFactory(object):
    """
    a factory object to generate quivers
    """

    def __init__(self, classifier_data, avg_rate_threshold=DEFAULT_AVG_RATE_THRESHOLD):
        self.classifier_data = classifier_data
        self.avg_rate_threshold = avg_rate_threshold

        ### divide I/O if specified
        if classifier_data.kwargs.get("direct", False): ### specifically skip dividing I/O
            self.divide_io = False
        elif ("window" in classifier_data.kwargs) and ("cache" in classifier_data.kwargs or "stride" in classifier_data.kwargs):
            self.divide_io = True
        else: ### missing needed classifier data kwargs to split sensibly
            self.divide_io = False

        if self.divide_io:
            self._classifier_datum = {}
            columns = classifier_data.columns

            ### divide I/O based on stride if available, default to cache segment boundaries otherwise
            if ("stride" in classifier_data.kwargs):
                file_segs = sorted(utils.split_segments_by_stride(classifier_data.start, classifier_data.end, classifier_data.segs, classifier_data.kwargs['stride']))
            else:
                file_segs = sorted(utils.cache2unique_segments(classifier_data.kwargs['cache']))

            ### generate a classifier data for each file span
            classifier_data_kwargs = dict(classifier_data.kwargs.items())
            for file_seg in file_segs:
                segs = utils.segments_intersection(classifier_data.segs, segmentlist([file_seg]))

                ### filter cache if classifier data uses one
                if ("cache" in classifier_data.kwargs):
                    classifier_data_kwargs['cache'] = [entry for entry in classifier_data.kwargs['cache'] if segs.intersects_segment(entry.segment)]

                self._classifier_datum[file_seg] = classifier_data.__class__(
                    float(file_seg[0]),
                    float(file_seg[1]),
                    segs=segs,
                    columns=columns,
                    **classifier_data_kwargs
                )

    def labeled(self, target_times, random_times):
        """
        construct a quiver based on classifier_data at these times
            target_times -> label=1.
            random_times -> label=0.
        """
        return self(
            np.concatenate((target_times, random_times)),
            labels=[1.]*len(target_times) + [0.]*len(random_times),
        )

    def unlabeled(self, dt=DEFAULT_DT, t_offset=0, segs=None):
        """
        construct a quiver based on classifier_data at a regular time interval dt.
        """
        assert t_offset <= dt, 'populate t_offset needs to be <= dt'
        if segs is None:
            segs = self.classifier_data.segs
        times = np.concatenate(tuple(utils.segs2times(segs, dt))) + t_offset
        return self(times)

    def __call__(self, times, labels=None):
        """
        a helper function to standardize how we assemble bits of quivers
        """
        if labels is None: ### set up labels if they aren't passed
            labels = [None]*len(times)
        else:
            assert len(labels)==len(times), 'labels and times must have the same length!'

        pairs = None ### tuples of (gps, label, classifier_data) later used to instantiate FeatureVectors
                     ### this may be overwritten within the following conditionals, otherwise we default to "standard behavior"

        ### choose whether to create a regular or dense quiver based on feature vector density
        if len(times) > 1:
            time_seg = utils.times2segments(times, 0.001).extent() ### small window just to estimate livetime
            dense = len(times) >= utils.livetime(segmentlist([time_seg])) * self.avg_rate_threshold
        else: ### can't predict density from only zero or one input times, use a regular quiver
            dense = False

        ### choose quiver type based on feature vector density
        if dense:
            quiver = features.DenseQuiver()
        else:
            quiver = features.Quiver()

        ### if specified, create feature vectors from classifier data directly and skip all the fancy stuff
        if not self.divide_io:
            pairs = [(gps, label, self.classifier_data) for gps, label in zip(times, labels)]

        ### else, determine whether to create a single umbrella based on feature vector density
        elif dense: ### single umbrella for all vectors
            time_segs = utils.times2segments(times, self.classifier_data.kwargs['window'])
            file_segs = list(set(itertools.chain(*[utils.find_unique_segs(seg, sorted(self._classifier_datum.keys())) for seg in time_segs])))
            relevant_datum = [self._classifier_datum[file_seg] for file_seg in sorted(file_segs)]
            relevant_segs = utils.segments_intersection(time_segs, utils.segments_union(*[cd.segs for cd in relevant_datum]))

            umbrella = triggers.UmbrellaClassifierData(
                min([cd.start for cd in relevant_datum]),
                max([cd.end for cd in relevant_datum]),
                segs=relevant_segs,
                columns=self.classifier_data.columns,
                **self.classifier_data.kwargs
            )
            umbrella.extend_children(relevant_datum)

            pairs = []
            for gps, label in zip(times, labels):
                pairs.append((gps, label, umbrella))

        else: ### create an umbrella for each feature vector
            ### condition on whether we filter segments around times
            if not self.classifier_data.kwargs.get('do_not_window', False):
                window_segs = utils.times2segments(times, self.classifier_data.kwargs['window'])

            pairs = []
            for gps, label in zip(times, labels):
                ### find relevant classifier datum based on gps window
                seg = segment(gps-self.classifier_data.kwargs['window'], gps+self.classifier_data.kwargs['window'])
                target_segs = utils.find_unique_segs(seg, sorted(self._classifier_datum.keys()))
                if len(target_segs) == 0: ### no data for this gps window
                    continue

                if len(target_segs) == 1:
                    umbrella_segs = self._classifier_datum[target_segs[0]].segs

                else: ### gps window spans over two classifier datum
                    umbrella_segs = self._classifier_datum[target_segs[0]].segs | self._classifier_datum[target_segs[1]].segs

                if not self.classifier_data.kwargs.get('do_not_window', False):
                    umbrella_segs = utils.segments_intersection(window_segs, umbrella_segs)

                umbrella = triggers.UmbrellaClassifierData(
                    target_segs[0][0],
                    target_segs[-1][1],
                    segs=umbrella_segs,
                    columns=self.classifier_data.columns,
                    **self.classifier_data.kwargs
                )

                ### add classifier datum to umbrella
                for target_seg in target_segs:
                    umbrella.extend_child(self._classifier_datum[target_seg])

                pairs.append((gps, label, umbrella))

        ### add in all the pairs defined above to populate quiver
        for gps, label, obj in pairs:
            quiver.append(features.FeatureVector(gps, label, obj))

        ### set quiver properties accordingly
        quiver.start = self.classifier_data.start
        quiver.end = self.classifier_data.end
        quiver.segs = self.classifier_data.segs

        ### if dense quiver, also set the classifier data property
        ### NOTE: by careful construction of pairs, all feature vectors share the same classifier data
        if dense:
            quiver.classifier_data = quiver[0].classifier_data

        return quiver
