__description__ = "a module to house all feature extraction and data pre-processing algorithms"
__author__ = "Reed Essick (reed.essick@ligo.org), Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

from collections import defaultdict
from collections import deque
from future.utils import string_types

import numpy as np
import multiprocessing as mp
import threading

import time as TIME

### non-standard libraries
from ligo.segments import segment, segmentlist

from . import utils
from . import exceptions

#-------------------------------------------------

DEFAULT_NUM_PROC = 1 ### number of processes used when parallelizing Quiver.vectorize, etc
DEFAULT_PARALLELIZATION= 'multiprocessing'
DEFAULT_MULTIPROCESSING_SLEEP = 0.001 ### how long we sleep between checking that mp.map_async results have finished
DEFAULT_MULTIPROCESSING_CHUNKSIZE = None

DEFAULT_TIME_NAME = 'time'
DEFAULT_DELTA_TIME = np.infty
DEFAULT_COLUMN_VALUE = -np.infty

DEFAULT_MAX_STRIDE = np.infty
DEFAULT_MAX_SAMPLES = np.infty

#-------------------------------------------------
### feature naming utility functions

FEATURE_NAME_TEMPLATE = "%s-%s"
def feature_name(channel, col):
    return FEATURE_NAME_TEMPLATE%(channel, col)

#-------------------------------------------------
### objects that downselect lists of triggers to a preferred trigger

class Select(object):
    """
    an object that downselects a list of triggers to a prefered one, otherwise returns a default value
    """
    _required_kwargs = []

    def __init__(self, default=DEFAULT_COLUMN_VALUE, **kwargs):
        self._default = default
        for kwarg in self._required_kwargs:
            assert kwarg in kwargs, 'required kwarg=%s missing'%kwarg
        self._kwargs = kwargs

    @property
    def default(self):
        return self._default

    def __call__(self, trgs, gps):
        raise NotImplementedError

class SelectLoudest(Select):
    """
    take the loudest thing within some time window
    if there is no trigger within the window, return None
    """
    _required_kwargs = ['window', 'time', 'significance']

    @property
    def window(self):
        return self._kwargs['window']

    @property
    def time(self):
        return self._kwargs['time']

    @property
    def significance(self):
        return self._kwargs['significance']

    def __call__(self, trgs, gps):
        if isinstance(gps, np.ndarray):
            ### time order
            sort_idx = np.argsort(gps)
            sort_gps = gps[sort_idx]

            ### set defaults
            loudest = np.empty(len(gps), dtype=trgs.dtype)
            loudest[:] = self.default

            ### search
            for trg in trgs:
                start_idx = np.searchsorted(sort_gps, trg[self.time] - self.window, side='left')
                end_idx = np.searchsorted(sort_gps, trg[self.time] + self.window, side='right')
                loudest[start_idx:end_idx] = np.where(
                    trg[self.significance] > loudest[start_idx:end_idx][self.significance],
                    trg,
                    loudest[start_idx:end_idx],
                )

            ### format and return
            unsort_idx = np.argsort(sort_idx)
            return loudest[unsort_idx]

        else:
            trgs = trgs[np.abs(trgs[self._kwargs['time']]-gps)<=self._kwargs['window']]
            if len(trgs):
                ### do the following backflip so that we return a structured array and not a "void type"
                return tuple(trgs[trgs[self._kwargs['significance']].argmax()]) ### return a tuple so index assignment works correctly

            else:
                return self._default ### should get mapped into index assignment correctly

class SelectPointy(Select):
    """
    take the thing with the smallest pointy-poisson p-value
    """
    _required_kwargs = ['window', 'time', 'significance']

    def __call__(self, trgs, gps):
        raise NotImplementedError

#-------------------------------------------------
### objects that perform xformations on an individual feature/vector basis

class ColumnMap(object):
    """
    manages transformations of columns on an individual feature/FeatureVector basis
    useful when the same transformation is applied to the same column from many channels

    Note: because I'm storing references to functions, this object will not be pickle-able...
    """
    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def __call__(self, col, x, default_x):
        if col in self._kwargs:
            transform_col, template, default = self._kwargs[col]
            return transform_col(x) if x is not default_x else default, template%col

        else:
            return x, col

class TimeMap(ColumnMap):
    """
    extends column map specifically to transform raw timestamps to relative timestamps
    """
    def __init__(self, gps, time=DEFAULT_TIME_NAME, default_delta_time=DEFAULT_DELTA_TIME):
        ColumnMap.__init__(self, **{time:((lambda x: x-gps), 'delta%s', default_delta_time)}) ### note, I go through the **kwargs backflip so I can dynamically assign the "time" name as a key in kwargs

#-------------------------------------------------
### objects that whiten features from quiver.vectorize and manage the xformations for us

class Whitener(object):
    """
    an object that manages whitening tables from quiver.vectorize
    remembers the xformations and serves as a utility for classifiers to map back and forth
    """

    def __init__(self):
        self.means = defaultdict(float)
        self.stdvs = defaultdict(float)

    def train(self, data, names):
        """
        assumes data.shape = (Nsamples, Nfeatures)
        """
        Nsamp, Nnames = data.shape
        assert Nnames==len(names), 'incompatible number of features/columns'

        for i, name in enumerate(names):
            self.means[name] = np.mean(data[:,i])
            s = np.std(data[:,i])
            self.stdvs[name] = s if s!=0 else 1.0

    def whiten(self, data, names):
        """
        map colored data into whitened data with xformation stored in this object
        assumes data.shape = (Nsamples, Nfeatures)

        modifies data in place, but also returns a reference
        """
        Nsamp, Nnames = data.shape
        assert Nnames==len(names), 'incompatible number of features/columns'

        means, stdvs = self._2array(names)

        ones = np.ones(Nsamp)
        data -= np.outer(ones, means)
        data /= np.outer(ones, stdvs)

        return data

    def color(self, data, names):
        """
        the inverse of whiten
        assumes data.shape = (Nsamples, Nfeatures)

        modifies data in place, but also returns a reference
        """
        Nsamp, Nnames = data.shape
        assert Nnames==len(names), 'incompatible number of features/columns'

        means, stdvs = self._2array(names)

        ones = np.ones(Nsamp)
        data *= np.outer(ones, stdvs)
        data += np.outer(ones, means)

        return data

    def _2array(self, names):
        Nnames = len(names)
        means = np.empty(Nnames, dtype='float')
        stdvs = np.empty(Nnames, dtype='float')
        for i, name in enumerate(names):
            if name not in self.means:
                raise KeyError('whitener has not been trained on data for name=%s'%name)
            means[i] = self.means[name]
            stdvs[i] = self.stdvs[name]
        return means, stdvs

#-------------------------------------------------

class FeatureVector(object):
    """
    an object that represents the feature vectors used in machine learning
    retains a pointer to a ClassifierData object for querying, etc
    While the ClassifierData objects do not know about labels (for supervised learning), FeatureVector objects do
    """

    def __init__(self, gps, label, classifier_data, **kwargs):
        self._gps = gps      ### the time-stamp associated with this vector
        self._label = label  ### the label associated with this vector -> can be a floating point number if it's something like p(glitch)
        self._rank = None    ### the rank associated with this vector
                             ### will be modified by ClassifierObjects through calls to evaluate
        self._model_hash = None ### records which model was used to evaluate this vector

        self.classifier_data = classifier_data   ### reference to data; used to extract features as needed

    @property
    def gps(self):
        return self._gps

    @property
    def label(self):
        return self._label

    @property
    def rank(self):
#        NOTE: we comment out this safety check because we want to be able to write unevaluated vectors to disk, which require rank to be accessbile
#        if not self.is_evaluated():
#            raise exceptions.UnevaluatedError('FeatureVector must be evaluated before rank can be determined (i.e.: call set_rank first)')
        return self._rank

    @rank.setter
    def rank(self, rank):
        """
        will be called by ClassifierObjects within their evaluate() methods
        """
        self._rank = rank

    @property
    def hash(self):
        return self._model_hash

    @hash.setter
    def hash(self, new):
        self._model_hash = new

    def is_evaluated(self):
        return self._rank != None

    def vectorize(self, channels, select, time=DEFAULT_TIME_NAME, default_delta_time=DEFAULT_DELTA_TIME):
        """
        actually compute the internal vector representation based on classifier_data
        """
        data, names = self.extract_data(channels, select)
        return self.transform_data(data, names, select.default, time=time, default_delta_time=default_delta_time)

    def extract_data(self, channels, select):
        """
        extract the relevant data from triggers obtained via self.classifier_data

        NOTE: does not return a structured array. Instead returns
            data, names
        """
        trgs = self.classifier_data.triggers(channels=channels) ### do I/O only once
        Ncol = len(self.classifier_data.columns)

        ### generate place holders
        data = np.empty(len(channels)*Ncol, dtype='float') ### a place-holder which we fill in below
        ### iterate over channels and select the best triggers, filling in default values as needed
        for i, channel in enumerate(channels):
            data[i*Ncol:(i+1)*Ncol] = select(trgs[channel], self.gps) ### NOTE: assumes ordering is consistent, which should be true

        return data, self.names(channels, select)

    def transform_data(self, data, names, select_default, time=DEFAULT_TIME_NAME, default_delta_time=DEFAULT_DELTA_TIME):
        """
        transform the data extracted via self.extract_data into the final vector
        """
        ### set up place-holder
        n = len(data)

        dtype = []
        column_map = TimeMap( ### NOTE: hard-coded to a TimeMap here. 
            self.gps,         ### If we want to generalize to other transformations, we should define subclasses of FeatureVector
            time=time,        ### or just pass **kwargs to a "ColumnMapFactory"?
            default_delta_time=default_delta_time,
        )

        for i in range(n):
            chan, col = names[i]
            data[i], col = column_map(col, data[i], select_default) ### we pass the default value so column_map can identify when data was missing
            dtype.append(feature_name(chan, col))

        return data, dtype

    def names(self, channels, select):
        return [(channel, col) for channel in channels for col in self.classifier_data.columns]

    def features(self, channels, select, time='time'):
        column_map = TimeMap(0, time=time)
        #define dummy values for data and select_default so we can pass them to column_map without issue
        dummy_data = 1
        dummy_select_default = 0
        return [feature_name(channel, column_map(col, dummy_data, dummy_select_default)[1]) for channel, col in self.names(channels, select)]

#---

class Quiver(deque):
    """
    an extension of a list which is meant to hold FeatureVectors and knows how to perform a few basic methods on those
    assumes all elements will be instances of FeatureVectors
    """
    _start = None ### defines the provinence for this quiver
    _end = None   ### FeatureVectors within this quiver should only fall within these start, end times
    _segs = []

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, start):
        self._start = start
        if self._segs:
            self.segs = utils.segments_intersection([(self._start, self._end)], self._segs)

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end):
        self._end = end
        if self._segs:
            self.segs = utils.segments_intersection([(self._start, self._end)], self._segs)

    @property
    def segs(self):
        return self._segs

    @segs.setter
    def segs(self, segs):
        """updates segments and checks coverage
        """
        if len(segs): ### at least one segment
            if self.start is None:
                self.start = segs[0][0] ### set start time to match segments

            if self.end is None:
                self.end = segs[-1][1] ### set end time to match segments

            assert utils.livetime(utils.segments_intersection(segmentlist([segment(self.start,self.end)]), segs)) == utils.livetime(segs), '''\
segments must be a subset of [[start=%.3f, end=%.3f]]; requested %s'''%(self.start, self.end, segs)
            self._segs = segs

        else: ### nothing in these segments
            self._segs = segs

    def check_coverage(self):
        """
        check to make sure all the FeatureVectors within this Quiver are within the segments
        """
        if not self.segs: ### segments have not been set
            raise AttributeError('please call Quiver.segs before check_coverage(), otherwise check_coverage() is ill defined')

        for vector in self:
            assert utils.time_in_segments(vector.gps, self.segs), 'vector(%.6f) not in segments %s!'%(vector.gps, self.segs)

    def filter(self, segs=None):
        """updates segments and filters out vectors that are not included, returning a quiver of the removed vectors
        """
        if segs is None:
            segs = self._segs
        else:
            self.segs = segs
        removed = Quiver()
        removed.start = self.start
        removed.end = self.end
        removed.segs = utils.remove_segments(segmentlist([segment(self.start, self.end)]), self.segs)
        if utils.livetime(removed.segs): ### there is something to do!
            N = len(self)
            for i in range(N): ### iterate over everything currently in self
                v = self.popleft()
                if utils.time_in_segments(v.gps, segs):
                    self.append(v)
                else:
                    removed.append(v)
        return removed

    def vectors2classes(self):
        '''
        this only makes sense if labels are integers: 0, 1
        one should not use this if labels are floats or something
        '''
        gch = Quiver()
        cln = Quiver()
        for q in [gch, cln]:
            q.start = self.start
            q.end = self.end
            q.segs = segmentlist(self.segs)

        for vector in self:
            if vector.label==1:
                gch.append(vector)
            elif vector.label==0:
                cln.append(vector)

        return gch, cln

    def vectors2num(self):
        """
        compute the number of glitch and clean samples in the feature set
        do this by counting their labels

        return num_gch, num_cln
        """
        labels = [vector.label for vector in self if vector.label != None]
        num_glitch = np.sum(labels)
        num_clean = len(labels) - num_glitch ### axiomatic because of how we define labels in 2-class classification

        return num_glitch, num_clean

    @property
    def labels(self):
        return np.array([v.label for v in self])

    @property
    def ranks(self):
        return np.array([v.rank for v in self])

    @property
    def hashes(self):
        return np.array([v.hash for v in self])

    @property
    def gpss(self):
        return np.array([v.gps for v in self])

    def apply_ranks(self, ranks, hashes=None):
        """
        apply a list of ranks to the vectors in this quiver. Assumes ordering of ranks is the same as the ordering of vectors
        """
        assert len(ranks)==len(self), 'ranks must have the same length as self'
        for r, v in zip(ranks, self):
            v.rank = r
        if hashes is not None:
            if isinstance(hashes, string_types): ### a single hash
                for v in self:
                    v.hash = hashes
            else:
                assert len(hashes)==len(self), 'hashes must be a single string or a list with the same length as self'
                for h, v in zip(hashes, self):
                    v.hash = h

    def copy(self):
        """
        return a copy of self so that clients can manipulate a quiver with tings like remove without mucking up shared references
        """
        new = Quiver()
        new.start = self.start
        new.end = self.end
        new.segs = segmentlist(self.segs)
        for vector in self:
            new.append(vector)
        return new

    def remove_quiver(self, quiver):
        """
        removes all the feature_vectors in vectors2remove from the current quiver
        does not assume any particular sorting
        WARNING: modifies object in place!
        """
        for vector in quiver:
            self.remove(vector) ### NOTE: will raise a ValueError if vector is not in self

    def append_quiver(self, other):
        """
        a convenience method for adding vectors from another quiver into this one
        """
        self.extend(other)

        self.start = min(self.start, other.start)
        self.end = max(self.end, other.end)
        self.segs = utils.segments_union(self.segs, other.segs)

    def flush(self, max_stride=DEFAULT_MAX_STRIDE, max_samples=DEFAULT_MAX_SAMPLES):
        self.start = max(self.start, self.end-max_stride) ### automatically updates segments
        removed = self.filter() ### filter using the internal segments
        while len(self) > max_samples: ### remove more samples as necessary
            removed.appendleft(self.popleft()) ### assumes implicit time ordering (we throw away old stuff first), but that's not strictly enforced
        return removed

    def vectorize(self, channels, select, num_proc=DEFAULT_NUM_PROC, io_parallelization=DEFAULT_PARALLELIZATION, cpu_parallelization=DEFAULT_PARALLELIZATION, **kwargs):
        """
        compute a label of vectorized representations as one big table/2-d array
        this should be suitable for input into things like sk_learn MLA libraries
        i.e. we hand our classifier objects a quiver, and the quiver knows how to format itself nicely

        NOTE: if num_proc!=1, we do this in parallel
        """
        assert len(self), 'must have at least one FeatureVector to vectorize a quiver'

        if num_proc==1:
            self._single_io(channels, **kwargs)
            return self._single_cpu(channels, select, **kwargs)

        else:
            try:
                if (io_parallelization=='multiprocessing') or (cpu_parallelization=='multiprocessing'):
                    pool = mp.Pool(num_proc) ### start a pool of this number of workers

                if (io_parallelization=='threading') or (cpu_parallelization=='threading'):
                    semaphore = threading.Semaphore(value=num_proc)

                ### parallelize I/O queries
                if io_parallelization is None:
                    self._single_io(channels, **kwargs)

                elif io_parallelization=='multiprocessing':
                    self._multiprocessing_io(pool, channels, **kwargs)

                elif io_parallelization=='threading':
                    self._threading_io(semaphore, channels, **kwargs)

                else:
                    raise ValueError('io_parallelization=%s not understood'%parallelization)

                ### parallelize construction of feature vectors
                if cpu_parallelization is None:
                    return self._single_cpu(channels, select, **kwargs)

                elif cpu_parallelization=='multiprocessing':
                    return self._multiprocessing_cpu(pool, channels, select, **kwargs)

                elif cpu_parallelization=='threading':
                    return self._threading_cpu(semaphore, channels, select, **kwargs)

                else:
                    raise ValueError('cpu_parallelization=%s not understood'%parallelization)

            except Exception as e:
                if (io_parallelization=='multiprocessing') or (cpu_parallelization=='multiprocessing'):
                    pool.terminate() ### kill all workers
                raise ### re-raise the Exception

    def _single_io(self, channels, **kwargs):
        """
        helper function for vectorize called when num_proc==1
        """
        for data in self._find_uncached_children(channels):
            data.triggers(channels)

    def _single_cpu(self, channels, select, **kwargs):
        """
        helper function for vectorize called when num_proc==1
        """
        data = []
        ### assumes names are the same for all vectors constructed... a bit fragile
        features = self[0].features(channels, select)
        for vector in self: ### assumes all data will report in the same order... a bit fragile
            data.append(vector.vectorize(channels, select, **kwargs)[0]) ### NOTE: pass kwargs so one quiver declaration can support multilple FeatureVector sublcasses,
                                                                      ### although it should be used with only one type of FeatureVector at a time...
        return np.array(data), features

    def _find_uncached_children(self, channels):
        """
        helper function that rapidly determines which classifier_data objects attached to vectors do not have all required data cached already
        """
        ### get a set of classifier_datas
        datas = set()
        for v in self:
            for child in v.classifier_data.children:
                datas.add(child)

        # iterate over the unique set so we don't repeat checks to see if data is already cached
        uncached = []
        for child in datas:
            ### NOTE: directly calling child._data.has_key(chan) is about twice as fast as calling child.is_cached
            ### this is almost certainly due to the extra time to look-up the functions within the delegation, but we value
            ### the clarify of the code (and avoiding using "private" attributes) enough that we'll suffer through as is
            ### FOr reference, using has_key for ~2k channels would take 3e-4 sec, while calls to is_cached takes 6e-4 sec
            for chan in channels:
                if not child.is_cached(chan):
                    break
            else:
                continue
            uncached.append(child)

        return uncached

    def _multiprocessing_io(self, pool, channels, multiprocessing_sleep=DEFAULT_MULTIPROCESSING_SLEEP, multiprocessing_chunksize=DEFAULT_MULTIPROCESSING_CHUNKSIZE, **kwargs):
        """
        helper function for vectorize that parallelizes using multiprocessing
        """
        ### get a set of classifier_datas that need to perform I/O
        uncached = self._find_uncached_children(channels)
        if uncached: ### need to fetch info for at least 1 object
            args = (channels,)
#                             this call will block until ready, but each job is run in parallel
#            for data, ans in zip(uncached, pool.map(_multiprocessing_fetch_triggers, ((obj, args, kwargs) for obj in uncached))):
#                data._data = ans ### map the parallel results onto the original objects

        ### NOTE: we do this so that zombie processes are automatically cleaned up instead of blocking until this returns
            result = pool.map_async(_multiprocessing_fetch_triggers, ((obj, args, kwargs) for obj in uncached), chunksize=multiprocessing_chunksize)
            while not result.ready():
                TIME.sleep(multiprocessing_sleep)
            if not result.successful():
                raise RuntimeError('call to multiprocessing.pool.map_async failed')

            for data, ans in zip(uncached, result.get()):
                data._data = ans ### map the parallel results onto the original objects

    def _multiprocessing_cpu(self, pool, channels, select, multiprocessing_sleep=DEFAULT_MULTIPROCESSING_SLEEP, multiprocessing_chunksize=DEFAULT_MULTIPROCESSING_CHUNKSIZE, **kwargs):
        """
        helper function for vectorize that parallelizes using multiprocessing
        """
        ### parallelize agglomeration into vectors
        features = self[0].features(channels, select)
        args = (channels, select)

#        data = pool.map(_multiprocessing_vectorize, ((v, args, kwargs) for v in self))
#        return np.array(data), features

        ### NOTE: we do this so that zombie processes are automatically cleaned up instead of blocking until this returns
        result = pool.map_async(_multiprocessing_vectorize, ((v, args, kwargs) for v in self), chunksize=multiprocessing_chunksize)
        while not result.ready():
            TIME.sleep(multiprocessing_sleep)
        if not result.successful():
            raise RuntimeError('call to multiprocessing.pool.map_async failed')

        return np.array(result.get()), features

    def _threading_io(self, semaphore, channels, **kwargs):
        """
        helper function for vectorize that parallelizes using threading
        multithreading may be slightly more dangerous than multiprocessing because of race conditions, etc, but we may be able to save on communcation costs by sharing memory
        """
        ### get a set of classifier_datas that need to perform I/O
        uncached = self._find_uncached_children(channels)

        if uncached: ### need to fetch info for at least 1 object
            args=(channels,)
            # launch a bunch of threads
            threads = []
            for data in uncached:
                thread = threading.Thread(target=_threading_fetch_triggers, args=(semaphore, [data], args, kwargs))
                thread.start()
                threads.append(thread)

            # wait for them all to complete. We might be able to do this with a barrier
            while len(threads):
                threads.pop(0).join()

    def _threading_cpu(self, semaphore, channels, select, **kwargs):
        """
        helper function for vectorize that parallelizes using threading
        """
        ### parallelize agglomeration into vectors
        names = self[0].names(channels, select)
        args = (channels, select)
        data = np.empty((len(self), len(names)), dtype='float') ### set up shared resource
        threads = []
        for ind, v in enumerate(self):
            thread = threading.Thread(target=_threading_vectorize, args=(semaphore, data, ind, [v], args, kwargs))
            thread.start()
            threads.append(thread)

        # wati for them all to complete. Again, we might be able to use a barrier here
        while len(threads):
            threads.pop(0).join()

        return data, names

class DenseQuiver(Quiver):
    """
    an extension of a quiver which is more suited to storing and handling densely spaced feature vectors.
    assumes all elements will be instances of FeatureVectors and that all FeatureVectors share the same ClassifierData.
    """
    _classifier_data = None

    @property
    def classifier_data(self):
        return self._classifier_data

    @classifier_data.setter
    def classifier_data(self, classifier_data):
        self._classifier_data = classifier_data
        for v in self:
            v.classifier_data = self.classifier_data

    def vectorize(self, channels, select, time=DEFAULT_TIME_NAME, default_delta_time=DEFAULT_DELTA_TIME, **kwargs):
        """
        compute a label of vectorized representations as one big table/2-d array
        this should be suitable for input into things like sk_learn MLA libraries
        i.e. we hand our classifier objects a quiver, and the quiver knows how to format itself nicely
        """
        assert len(self), 'must have at least one FeatureVector to vectorize a quiver'
        data, names = self.extract_data(channels, select, time=time)
        return self.transform_data(data, names, select.default, time=time, default_delta_time=default_delta_time)

    def extract_data(self, channels, select, time=DEFAULT_TIME_NAME):
        """
        extract the relevant data from triggers obtained via self.classifier_data

        NOTE: does not return a structured array. Instead returns
            data, names
        """
        ### define max extent for triggers to minimize I/O
        gps_segs = segmentlist([segment(min(self.gpss)-select.window, max(self.gpss)+select.window)])
        relevant_segs = utils.segments_intersection(gps_segs, self.segs)

        ### grab triggers
        trgs = self.classifier_data.triggers(channels=channels, time=time, segs=relevant_segs) ### do I/O only once
        Ncol = len(self.classifier_data.columns)

        ### generate place holders
        data = np.empty((len(self), len(channels)*Ncol), dtype='float') ### a place-holder which we fill in below
        ### iterate over channels and select the best triggers, filling in default values as needed
        for i, channel in enumerate(channels):
            data[:, i*Ncol:(i+1)*Ncol] = select(trgs[channel], self.gpss).view((float, Ncol)) ### NOTE: assumes ordering is consistent, which should be true

        return data, self.names(channels, select)

    def transform_data(self, data, names, select_default, time=DEFAULT_TIME_NAME, default_delta_time=DEFAULT_DELTA_TIME):
        """
        transform the data extracted via self.extract_data into a quiver
        """
        dtype = []
        column_map = TimeMap( ### NOTE: hard-coded to a TimeMap here.
            self.gpss,        ### If we want to generalize to other transformations, we should define subclasses of FeatureVector
            time=time,        ### or just pass **kwargs to a "ColumnMapFactory"?
            default_delta_time=default_delta_time,
			)

        for i in range(len(names)):
            chan, col = names[i]
            data[:,i], col = column_map(col, data[:,i], select_default) ### we pass the default value so column_map can identify when data was missing
            dtype.append(feature_name(chan, col))

        return data, dtype

    def names(self, channels, select):
        return [(channel, col) for channel in channels for col in self.classifier_data.columns]

    def features(self, channels, select, time='time'):
        column_map = TimeMap(0, time=time)
        #define dummy values for data and select_default so we can pass them to column_map without issue
        dummy_data = 1
        dummy_select_default = 0
        return [feature_name(channel, column_map(col, dummy_data, dummy_select_default)[1]) for channel, col in self.names(channels, select)]

#-------------------------------------------------

@utils.unpack
def _multiprocessing_fetch_triggers(data, args, kwargs):
    '''
    a helper method used within multiprocessing parallelization to fetch triggers (do all the I/O first)
    needed because we can't pickle an instancemethod as the target function
    '''
    return data.triggers(**kwargs)

@utils.unpack
def _multiprocessing_vectorize(vector, args, kwargs):
    '''
    a helper method used within multiprocessing parallelization to vectorize each vector
    needed because we can't pickle an instancemethod as the target function
    '''
    return vector.vectorize(*args, **kwargs)[0]

#------------------------

def _threading_fetch_triggers(semaphore, datas, args, kwargs):
    '''
    helper method used within threading parallelization to fetch triggers (do all the I/O first)
    '''
    with semaphore:
        for data in datas:
            data.triggers(*args, **kwargs)

def _threading_vectorize(semaphore, shared_array, global_ind, vectors, args, kwargs):
    '''
    helper method used within threading parallelization to vectorize each vector
    '''
    with semaphore:
        for ind, vector in enumerate(vectors):
            data, names = vector.vectorize(*args, **kwargs)
            shared_array[global_ind+ind] = data ### I almost certainly do need a lock here!
