__description__ = "a python module housing KW-based classifier datum"
__author__ = "Reed Essick (reed.essick@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

from collections import defaultdict
import glob
import os

import h5py
import numpy as np

from ligo.segments import segment, segmentlist

from ... import features
from ... import names
from ... import utils
from ... import hookimpl
from . import ClassifierData


#-------------------------------------------------
### classifier data implementations

def _kws2triggers(path, segments=None, columns=utils.KW_COLUMNS, bounds=None):
    if bounds is None:
        bounds = dict()
    if segments is None:
        segments = segmentlist([segment(-np.inf, +np.inf)]) ### grab everything

    data = []
    num_segs = len(segments)
    if num_segs == 0:
        return data

    segind = 0
    inds = [utils.KW_COL2IND[col] for col in columns] ### which indecies we extract from each line
    bounds = [(utils.KW_COL2IND[col], tup) for col, tup in bounds.items()]

    with open(path, 'r') as file_obj:
        start, end = segments[segind] ### which segment we're working with

        for line in file_obj:
            if line[0] in utils.ASCII_COMMENTS: ### commented
                continue

            fields = _kws_line2fields(line) ### parse the line
            t = fields[utils.KW_TIMEIND]

            # handle segment logic here in an attempt to speed things up (we assume sorting here)
            if t > end: ### trigger is too late, bump the segment
                segind += 1
                while segind < num_segs:
                    if t > segments[segind][1]:
                        segind += 1
                    else:
                        break
                else:
                    break ### did not find another segment, so we're done with this file
                start, end = segments[segind] # update definition of start, end

            if t < start: ### trigger is too early, but we still have segments, so go to next trigger
                continue

            ### now, start < t < end is guaranteed, so we add this trigger
            ### first, we filter by bounds
            for col, (m, M) in bounds:
                v = fields[col]
                if (v<m) or (M<v): ### outside the bounds, so we do not include this line
                    break
            else: ### no one broke, so we're good
                data.append( tuple(fields[inds]) )

    return data

def _kws_line2fields(line):
    return np.array([float(val) for val in line.strip().split()]) ### NOTE: this is faster than np.array(...).astype('float')

def _kwm2triggers(path, segments=None, columns=utils.KW_COLUMNS, channels=None, bounds=None):
    if bounds is None:
        bounds = dict()

    if segments is None:
        segments = segmentlist([segment(-np.inf, +np.inf)]) ### grab everything

    data = defaultdict(list)
    num_segs = len(segments)
    if num_segs == 0:
        return data

    segind = 0
    inds = [utils.KW_COL2IND[col] for col in columns] ### which indecies we extract from each line
    bounds = [(utils.KW_COL2IND[col], tup) for col, tup in bounds.items()]

    with open(path, 'r') as file_obj:
        start, end = segments[segind]

        for line in file_obj:
            if line[0] in utils.ASCII_COMMENTS:
                continue

            fields, chan = _kwm_line2fields(line)
            if channels and (chan not in channels): ### we don't care about this channel
                continue

            # we do care about this channel, so now we check segments
            t = fields[utils.KW_TIMEIND]

            # handle segment logic here in an attempt to speed things up (we assume sorting here)
            if t > end: ### trigger is too late, bump the segment
                segind += 1
                while segind < num_segs:
                    if t > segments[segind][1]:
                        segind += 1
                    else:
                        break
                else:
                    break ### did not find another segment, so we're done with this file
                start, end = segments[segind] # update definition of start, end

            if t < start: ### trigger is too early, but we still have segments, so go to next trigger
                continue

            ### now, start < t < end is guaranteed, so we add this segment
            for col, (m, M) in bounds:
                v = fields[col]
                if (v<m) or (M<v): ### outside the bounds
                    break
            else: ### no-one broke, so we add this
                data[chan].append( tuple(fields[inds]) )

    return data

def _kwm_line2fields(line):
    fields = line.strip().split()
    return np.array([float(val) for val in fields[:-1]]), fields[-1]

#-----------

class PredictiveKWSClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for KleineWelle triggers.
    expects triggers to be in single-channel files organized in a directory structure with a constant (known) stride
    Note: if we do not request specific channels, all discoverable channels will be returned
    """
    _default_columns = utils.KW_COLUMNS
    _allowed_columns = utils.KW_COLUMNS ### allow only KW columns
    _required_kwargs = ['rootdir', 'stride']

    def _retrieve_triggers(self, channels=None, bounds=None, verbose=False, **kwargs):
        """
        workhorse data discovery method for KWS files
        predicts directory names through expected directory structure and then looks for channel.trg files within each directory as needed
        we assume the following directory structure
            ${rootdir}/${gpsstart}_${gpsstop}/${channel}.trg
        """
        if bounds is None:
            bounds = dict()

        if verbose:
            print( '''\
searching for KWSingle triggers within
    rootdir : %(rootdir)s
with
    stride : %(stride)d'''%self.kwargs)
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print('within segments')
            for seg in self.segs:
                print('    %.3f - %.3f'%tuple(seg) )

        # extract things that will be used repeatedly
        stride = self.kwargs['stride']

        ### iterate through segments, attaching directories to each segment as needed
        template = names.start_end2dirtemplate(rootdir=self.kwargs['rootdir'])
        directories = defaultdict(segmentlist)
        for s, e in self.segs:
            S = (int(s)/stride)*stride ### integer division to get the directory starting point!
            while S < e:
                E = S+stride
                directory = template%(S, E)
                if os.path.exists(directory):
                    directories[directory].append(segment(s, min(e,E)))
                S = E ### advance by one stride

        ### now actually read in triggers
        data = defaultdict(list) ### data structure for query results
        if channels is not None:
            chan_paths = [(chan, names.kws_channel2filename(chan)) for chan in channels] ### define this for convenience later

        # iterate through directories that exist
        for directory in sorted(directories.keys()): ### make sure these are sorted
            if verbose:
                print( '    '+directory )

            relevant_segments = directories[directory] ### extract segments for this directory
            if verbose:
                for seg in relevant_segments:
                    print( '        %.3f - %.3f'%tuple(seg) )

            ### get relevant paths for this directory
            ### the resulting list is guaranteed to only contain paths that exist
            if channels is not None:
                paths = []
                for chan, chan_path in chan_paths:
                    path = os.path.join(directory, chan_path)
                    if os.path.exists(path):
                        paths.append((chan, path))

            else:
                paths = [(os.path.basename(path)[:-4], path) for path in glob.iglob(os.path.join(directory, '*.trg'))]

            # iterate through paths, reading in triggers for each path
            for chan, path in paths:
                if verbose:
                    print( '        '+path )
                data[chan] += _kws2triggers(
                    path,
                    segments=utils.segments_intersection(relevant_segments, self.segs),
                    columns=self.columns,
                    bounds=bounds, ### filter as we read the file
                )

        ### clean crap up
        dtype = [(key, 'float') for key in self.columns]
        for k, v in data.items(): ### cast everything into a structured array
            data[k] = np.array(v, dtype=dtype)

        if channels is not None:
            for chan in channels: ### fill in missing channels
                if not chan in data:
                    data[chan] = np.array([], dtype=dtype)

        return data


class KWSClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for KleineWelle triggers.
    expects triggers to be in single-channel files
    Note: if we do not request specific channels, all discoverable channels will be returned
    """
    _default_columns = utils.KW_COLUMNS
    _allowed_columns = utils.KW_COLUMNS ### allow only KW columns
    _required_kwargs = ['rootdir']

    def _retrieve_triggers(self, channels=None, bounds=None, verbose=False, **kwargs):
        """
        workhorse data discovery method for KWS files
        globs through expected directory structure only once and then looks for channel.trg files within each directory as needed
        we assume the following directory structure
            ${rootdir}/${gpsstart}_${gpsstop}/${channel}.trg
        """
        if bounds is None:
            bounds = dict()

        if verbose:
            print( '''\
searching for KWSingle triggers within
    rootdir : %(rootdir)s'''%self.kwargs )
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print( 'within segments' )
            for seg in self.segs:
                print( '    %.3f - %.3f'%tuple(seg) )

        data = defaultdict(list) ### data structure for query results

        if channels is not None:
            chan_paths = [(chan, names.kws_channel2filename(chan)) for chan in channels] ### define this for convenience later

        ### find which directories contain the stretches we want
        for directory in glob.iglob(os.path.join(self.kwargs['rootdir'], '*_*/')):
            if verbose:
                print( '    '+directory )

            dir_seg = segment(names.dir2start_end(directory))

            ### determine which segments have any overlap with this directory
            ### will save time when we filter triggers specifically from this directory
            relevant_segments = segmentlist([])
            for seg in self.segs: ### check for overlap with any segments
                if utils.segments_overlap(dir_seg, seg): ### non-zero overlap
                    relevant_segments.append( seg )

            if not relevant_segments: ### no overlap
                continue

            if verbose:
                for seg in relevant_segments:
                    print( '        %.3f - %.3f'%tuple(seg) )

            if channels is not None: ### only load a specific list
                paths = []
                for chan, chan_path in chan_paths:
                    path = os.path.join(directory, chan_path) ### predict the exact path and
                                                              ### require an exact match
                    if os.path.exists(path):
                        paths.append((chan, path))

            else: ### load everything you can find
                paths = [(os.path.basename(path)[:-4], path) for path in glob.iglob(os.path.join(directory, '*.trg'))]

            for chan, path in paths:
                if verbose:
                    print( '        '+path )
                data[chan] += _kws2triggers(
                    path,
                    segments=utils.segments_intersection(relevant_segments, self.segs),
                    columns=self.columns,
                    bounds=bounds, ### filter as we read the file
                )

        dtype = [(key, 'float') for key in self.columns]
        for k, v in data.items(): ### cast everything into a structured array
            data[k] = np.array(v, dtype=dtype)

        if channels is not None:
            for chan in channels: ### fill in missing channels
                if not chan in data:
                    data[chan] = np.array([], dtype=dtype)

        return data


class PredictiveKWMClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for KleineWelle triggers.
    expects triggers to be in multi-channel files with a known stride

    Note, if we do not request any specific channel(s), all discoverable channels will be returned
    """
    _default_columns = utils.KW_COLUMNS
    _allowed_columns = utils.KW_COLUMNS ### allow only KW columns
    _required_kwargs = ['rootdir', 'basename', 'stride']

    def _retrieve_triggers(self, channels=None, bounds=None, verbose=False, **kwargs):
        """
        read all relevant triggers from a single KWM file in at once to save on I/O
        we predict filenames and assume the following directory structure
            ${rootdir}/${basename}-${gpsMOD1e5}/${basename}-${start}-${dur}.trg
        """
        if bounds is None:
            bounds = dict()

        if verbose:
            print( '''\
searching for KWMultiple triggers within
    rootdir : %(rootdir)s
with
    basename : %(basename)s
    stride   : %(stride)d'''%self.kwargs )
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print( 'within segments' )
            for seg in self.segs:
                print( '    %.3f - %.3f'%tuple(seg) )

        ### pull out things referenced more than once
        stride = self.kwargs['stride']
        basename = self.kwargs['basename']

        ### iterate through directories, discovering which paths are relevant
        paths = defaultdict(segmentlist)

        template = names.start2dirMOD1e5_template(basename=self.kwargs['basename'], rootdir=self.kwargs['rootdir'])

        for s, e in self.segs:
            ints = int(s)
            Smod5 = (ints/100000) # integer division; hard-coded because this is what KWM always produces
            S = Smod5*100000

            while S < e: ### iterate over directories
                E = S+100000
                end = min(E, e) # the relevant end time for this directory
                SS = (max(ints, S)/stride)*stride ### the relevant starting point for this directory

                path_template = os.path.join(template%Smod5, names.kwm_stridebasename2filename(stride=stride, basename=basename))
                while SS < end: ### iterate over files in this directory
                    EE = SS+stride
                    path = path_template%SS

                    if os.path.exists(path):
                        paths[path].append(segment(max(SS, s), min(EE, e)))
                    SS = EE # bump file start time by one stride

                S = E # bump directory start time by one stride (100000%100000 = 1)

        ### iterate over paths and extract data
        data = defaultdict(list) ### data structure for query results
        for path in sorted(paths.keys()): ### load these in in order
            if verbose:
                print(path)

            relevant_segments = paths[path] ### extract the useful segments
            if verbose:
                for seg in relevant_segments:
                    print('    %.3f - %.3f'%tuple(seg))

            for channel, triggers in _kwm2triggers(
                    path,
                    segments=utils.segments_intersection(relevant_segments, self.segs),
                    columns=self.columns,
                    channels=channels,
                    bounds=bounds, ### filter as we read the file
                ).items():
                data[channel] += triggers

        dtype = [(key, 'float') for key in self.columns]
        for k, v in data.items(): ### cast everything into a structured array
            data[k] = np.array(v, dtype=dtype)

        if channels:
            for chan in channels: ### fill in missing channels
                if chan not in data:
                    data[chan] = np.array([], dtype=dtype)

        return data


class KWMClassifierData(ClassifierData):
    """
    an extension of ClassifierData specifically for KleineWelle triggers.
    expects triggers to be in multi-channel files

    Note, if we do not request any specific channel(s), all discoverable channels will be returned
    """
    _default_columns = utils.KW_COLUMNS
    _allowed_columns = utils.KW_COLUMNS ### allow only KW columns
    _required_kwargs = ['rootdir', 'basename']

    def _retrieve_triggers(self, channels=None, bounds=None, verbose=False, **kwargs):
        """
        read all relevant triggers from a single KWM file in at once to save on I/O
        we assume the following directory structure
            ${rootdir}/${basename}-${gpsMOD1e5}/${basename}-${start}-${dur}.trg
        """
        if bounds is None:
            bounds = dict()

        if verbose:
            print( '''\
searching for KWMultiple triggers within
    rootdir : %(rootdir)s
with
    basename : %(basename)s'''%self.kwargs )
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print( 'within segments' )
            for seg in self.segs:
                print( '    %.3f - %.3f'%tuple(seg) )

        data = defaultdict(list) ### data structure for query results

        basename_template = self.kwargs['basename']+'-*.trg' ### define this once, will be used repeatedly below

        ### find which directories contain the stretches we want
        for directory in glob.iglob(os.path.join(self.kwargs['rootdir'], self.kwargs['basename']+'-*/')):
            if verbose:
                print( '    '+directory )

            dir_seg = segment(names.dirMOD1e52start_end(directory))

            ### determine which segments are useful for this directory
            ### will save time later on
            relevant_segments = segmentlist([])
            for seg in self.segs: ### check for overlap with any segments
                if utils.segments_overlap(dir_seg, seg): ### non-zero overlap
                    relevant_segments.append( seg )

            if not relevant_segments: ### no overlap
                continue

            if verbose:
                for seg in relevant_segments:
                    print( '        %.3f - %.3f'%tuple(seg) )

            for trg_path in glob.iglob(os.path.join(directory, basename_template)): ### iterate over KWM files
                trg_seg = segment(names.path2start_end(trg_path, suffix='trg'))

                germane_segments = segmentlist([]) ### same idea as relevant segments, but just for this trg file
                for seg in relevant_segments:
                    if utils.segments_overlap(trg_seg, seg):
                        germane_segments.append( seg )

                if not germane_segments: ### no overlap
                    continue

                if verbose:
                    print( '        '+trg_path )

                for channel, triggers in _kwm2triggers(
                        trg_path,
                        segments=utils.segments_intersection(self.segs, germane_segments),
                        columns=self.columns,
                        channels=channels,
                        bounds=bounds, ### filter as we read the file
                    ).items():
                    data[channel] += triggers

        dtype = [(key, 'float') for key in self.columns]
        for k, v in data.items(): ### cast everything into a structured array
            data[k] = np.array(v, dtype=dtype)

        if channels:
            for chan in channels: ### fill in missing channels
                if chan not in data:
                    data[chan] = np.array([], dtype=dtype)

        return data


class KWMHDF5ClassifierData(ClassifierData):
    """
    An extension of ClassifierData specifically for extracting KleineWelle features from hdf5 files.

    We assume the following directory structure:
        ${rootdir}/${basename}/${basename}-${gpsMOD1e5}/${basename}-${start}-${dur}.h5

    """
    _default_columns = utils.KW_COLUMNS
    _allowed_columns = utils.KW_COLUMNS
    _required_kwargs = ['rootdir', 'basename']

    def __init__(self, start, end, segs=None, columns=None, **kwargs):
        ClassifierData.__init__(self, start, end, segs=segs, columns=columns, **kwargs)

        ### load cache, or create one if one is not provided
        populate_cache = False
        if 'cache' in self.kwargs and isinstance(self.kwargs['cache'], list):
            if self.kwargs['cache']:
                cache_extent = segment(self.kwargs['cache'][0].segment[0], self.kwargs['cache'][-1].segment[1])
                if utils.livetime(self.segs) > 0 and not (self.segs.extent() in cache_extent):
                    populate_cache = True ### zero livetime or cache doesn't cover span of classifier data
            else:
                populate_cache = True ### cache is empty
        elif 'cache' in self.kwargs:
            self.kwargs['cache'] = utils.file2cache(self.kwargs['rootdir'], self.kwargs['cache'])
        else:
            populate_cache = True

        if populate_cache:
            path = names.basename2gstlal_filedir(self.kwargs['basename'])
            filename = names.basename2gstlal_filename(self.kwargs['basename'])
            self.kwargs['cache'] = utils.path2cache(self.kwargs['rootdir'], os.path.join(path, filename))

        ### filter cache with segments
        self.kwargs['cache'] = [entry for entry in self.kwargs['cache'] if self.segs.intersects_segment(entry.segment)]

    def restrict_segments(self, new_segs, time=features.DEFAULT_TIME_NAME):
        ClassifierData.restrict_segments(self, new_segs, time=time) ### delegate to the parent
        self.kwargs['cache'] = [entry for entry in self.kwargs['cache'] if self.segs.intersects_segment(entry.segment)] ### restrict the cache

    def _retrieve_triggers(self, channels=None, bounds=None, verbose=False, **kwargs):
        """
        workhorse data discovery method for KleineWelle hdf5 files
        we assume the following directory structure
            ${rootdir}/${basename}/${basename}-${gpsMOD1e5}/${basename}-${start}-${dur}.h5
        """
        if bounds is None:
            bounds = dict()

        if verbose:
            print( '\n'.join([
                '',
                'searching for KW triggers within',
                '    rootdir : {rootdir}'.format(**self.kwargs),
                'with',
                '    basename : {basename}'.format(**self.kwargs),
                ])
            )
            if channels:
                print('for channels')
                for chan in channels:
                    print( '    '+chan )
            print( 'within segments' )
            for seg in self.segs:
                print( '    {:3f} - {:3f}'.format(*seg) )

        ### data structure and dtype for query results
        data = defaultdict(list)
        dtype = [(col, 'float') for col in self.columns]

        for entry in self.kwargs['cache']: ### iterate over hdf5 files

            ### determine which segments are relevant for this hdf5 file
            trg_seg = segment(names.path2start_end(entry.path, suffix='h5'))
            germane_segments = utils.segments_intersection(self.segs, segmentlist([trg_seg]))

            if verbose:
                print( '        '+entry.path )

            with h5py.File(entry.path, 'r') as file_obj:

                ### query only channels in file
                file_channels = set(file_obj.keys())
                if channels is not None:
                    channels2query = file_channels & set(channels)
                else:
                    channels2query = file_channels - set(self.channels)

                for channel in channels2query:

                    ### filter datasets with segments
                    datasets = list(set(file_obj[channel].keys()))
                    dset_segs = [segment(*utils.start_dur2start_end(dset, type_=float)) for dset in datasets]
                    relevant_datasets = [(dset, seg) for dset, seg in zip(datasets, dset_segs) if germane_segments.intersects_segment(seg)]

                    if relevant_datasets:
                        key = os.path.join(channel, relevant_datasets[0][0])
                        row_dtype = file_obj[key].dtype

                        ### read in all datasets associated with this channel
                        rows = []
                        for dataset, seg in relevant_datasets:
                            key = os.path.join(channel, dataset)
                            rows.append(np.array(file_obj[key])[self.columns])
                        rows = np.concatenate(rows, axis=0)

                        ### filter after reading everything from this file
                        rows = utils.filter_triggers(rows, bounds=bounds)
                        trimmed_times = utils.times_in_segments(rows[utils.KW_TIMECOL], germane_segments)
                        data[channel].append(rows[trimmed_times])

        ### stack datasets vertically
        for channel, datasets in data.items():
            data[channel] = np.concatenate(datasets, axis=0)

        ### fill in missing channels
        if channels:
            for channel in channels:
                if channel not in data:
                    data[channel] = np.array([], dtype=dtype)

        return data


#-------------------------------------------------
### plugin implementations

@hookimpl
def get_classifier_datum():
    return {
        "kw:single": KWSClassifierData,
        "kw:multi": KWMClassifierData,
        "kw:single-predictive": PredictiveKWSClassifierData,
        "kw:multi-predictive": PredictiveKWMClassifierData,
        "kw:multi-hdf5": KWMHDF5ClassifierData,
    }
