__description__ = "a python module housing omicron-based classifier datum"
__author__ = "Reed Essick (reed.essick@ligo.org), Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#-------------------------------------------------

from collections import defaultdict
import glob
import os

import h5py
import numpy as np

from ligo.segments import segment, segmentlist

from ... import names
from ... import utils
from ... import hookimpl
from . import ClassifierData


#-------------------------------------------------
### classifier data implementations

class OmicronClassifierData(ClassifierData):
    """an extension meant to read Omicron triggers off disk
    """
    _default_columns = utils.OMICRON_COLUMNS
    _allowed_columns = utils.OMICRON_COLUMNS
    _required_kwargs = ['instrument', 'rootdir']

    _suffix = 'h5'

    def _retrieve_triggers(self, channels=None, bounds=None, verbose=False, **kwargs):
        if bounds is None:
            bounds = dict()

        if channels is None:
            channels = [os.path.basename(directory) for directory in glob.iglob(os.path.join(self.kwargs['rootdir'], '*'))]

        data = defaultdict(list)
        dtype = [(key, 'float') for key in self.columns]

        ### find which directories contain the stretches we want
        for channel in channels:
            if verbose:
                print( 'processing: '+channel )
            template = names.channel2omicron_filename(self.kwargs['instrument'], channel, suffix=self._suffix)
            for directory in glob.iglob(os.path.join(self.kwargs['rootdir'], channel, '*')):
                if verbose:
                    print( '    '+directory )

                start = int(os.path.basename(directory))*100000
                end = start + 1000000
                dir_seg = segment(start, end)

                ### determine which segments are useful for this directory
                relevant_segments = segmentlist([])
                for seg in self.segs: ### check for overlap with any segments
                    if utils.segments_overlap(dir_seg, seg): ### non-zero overlap
                        relevant_segments.append( seg )

                if not relevant_segments: ### no overlap
                    continue

                if verbose:
                    for seg in relevant_segments:
                        print( '        %.3f - %.3f'%tuple(seg) )

                for trg_path in glob.iglob(os.path.join(directory, template)): ### iterate over KWM files
                    trg_seg = segment(names.path2start_end(trg_path, suffix=self._suffix))

                    germane_segments = segmentlist([]) ### same idea as relevant segments, but just for this trg file
                    for seg in relevant_segments:
                        if utils.segments_overlap(trg_seg, seg):
                            germane_segments.append( seg )

                    if not germane_segments: ### no overlap
                        continue

                    if verbose:
                        print( '        '+trg_path )

                    try:
                        with h5py.File(trg_path, 'r') as obj:
                            dataset = obj['triggers']
                            datum = np.array(dataset)[self.columns]
                    except OSError as e:
                        ### NOTE: Some production Omicron files are
                        ### found with insufficient permissions which
                        ### can cause problems. We can optionally
                        ### skip over these problematic files.
                        if not self.kwargs.get('skip_bad_files', False):
                            raise
                        print(f"unable to read {trg_path}, skipping...")
                        continue
                    else:
                        ### filter as data is read in by bounds and segments
                        datum = utils.filter_triggers(datum, bounds=bounds)
                        datum = datum[utils.times_in_segments(datum[utils.OMICRON_TIMECOL], germane_segments)]
                        if len(datum): ### there is something left to include
                            data[channel].append(datum)

        ### stack datasets vertically
        for channel, datasets in data.items():
            data[channel] = np.concatenate(datasets, axis=0)

        ### fill in missing channels
        if channels:
            for channel in channels:
                if channel not in data:
                    data[channel] = np.array([], dtype=dtype)

        return data


#-------------------------------------------------
### plugin implementations

@hookimpl
def get_classifier_datum():
    return {
        "omicron": OmicronClassifierData,
        "omicron:hdf5": OmicronClassifierData,
    }
